import pandas as pd
import json
import copy
from openpyxl.drawing.image import Image
from openpyxl import load_workbook
from openpyxl.styles import Font, PatternFill, Alignment

from .build_defaults import build_default_formats
from .options import Options
from .sheet import Sheet

class Build:
    """
    A class that gathers together crosstabs a user wants to use to build Excel, Powerpoint or Datasmoothie dashboards output. 

    Parameters
    ----------
      name: string
        Name for the dataset
    """

    def __init__(self, name=None, subtitle=None, default_dataset=None, table_of_contents=False, logo=None):
        self.name = name
        self.subtitle = subtitle
        self.default_dataset = default_dataset
        self.table_of_contents = table_of_contents
        self.sheets = []
        self.logo = logo
        self.font_name = None
        self.font_size = None
        self.index_options = {
            'header_color': 'ffffff',
            'body_color': 'ffffff',
            'link_color': '2A64C5'
        }
        self.options = Options(
            formats=copy.deepcopy(build_default_formats),
            parent=self
        )

    def add_sheet(self, name=None, banner='@'):
        if name is None:
            name = "Table {}".format(str(len(self.sheets)+1))
        sheet = Sheet(
            banner=banner, 
            default_dataset=self.default_dataset, 
            name=name,
            parent=self
            )
        self.sheets.append(sheet)
        return sheet

    def add_logo(self, path):
        self.logo = path

    def set_index_option(self, name, option):
        self.index_options[name] = option

    def save_excel(self, filename):
        dataframes_payload = []
        formats = []
        for sheet in self.sheets:
            df = sheet.build_dataframes(self.options)
            df = df.replace({'null':0})
            payload = json.loads(df.to_json(orient='split'))
            payload['index_names'] = df.index.names.copy()
            payload['column_names'] = df.columns.names.copy()
            dataframes_payload.append(payload)        
            formats.append(sheet.options.formats)

        self.default_dataset.build_excel_from_dataframes(filename=filename, dataframes=dataframes_payload, client_formats=formats)


        # wrapped in try because openpyxl is brittle
        try:
            heights = {
                'Question': 30,
                'Values': 70,
                'Test-IDs': 30
            }

            wb = load_workbook(filename)

            default_font = Font(name=self.font_name, size=self.font_size)
            
            # add table of contents
            if len(self.sheets)>1:
                offset_y = 12

                wb_sheet = wb.create_sheet('Contents', 0)
                wb_sheet.cell(row=offset_y, column=2, value='Contents')
                wb_sheet.column_dimensions['B'].width = 20
                wb_sheet.column_dimensions['C'].width = 70

                if self.name is not None:
                    title = wb_sheet.cell(7, 2, value=self.name)
                    title.font = Font(size="20")
                if self.subtitle is not None:
                    subtitle = wb_sheet.cell(8, 2, value=self.subtitle)
                    subtitle.font = Font(size="14")

                for index, sheet in enumerate(self.sheets):
                    wb_sheet.cell(row=index+1+offset_y, column=3, value=sheet.get_name())
                    link_value = '=HYPERLINK("#{}!A1", "Table {}")'.format(wb.sheetnames[index+1], index+1)
                    cell = wb_sheet.cell(row=index+1+offset_y, column=2, value=link_value)
                    cell.font = Font(color="2A64C5", underline="single")

                if self.logo is not None:
                    img = Image(self.logo)
                    wb_sheet.add_image(img, 'B2')

                for row_range in range(1, 10):
                    for col_range in range(1, 30):
                        color_cell = wb_sheet.cell(row_range, col_range)
                        color_cell.fill = PatternFill(start_color=self.index_options['header_color'], end_color=self.index_options['header_color'], fill_type="solid")

                # self.sheets tells us how many entries are in the index
                for row_range in range(10, len(self.sheets)+100):
                    for col_range in range(1, 30):
                        color_cell = wb_sheet.cell(row_range, col_range)
                        color_cell.fill = PatternFill(start_color=self.index_options['body_color'], end_color=self.index_options['body_color'], fill_type="solid")
                        #color_cell.font = default_font

                for index, sh in enumerate(self.sheets):
                    top_offset = sh.options.formats['offsets']['top']
                    wb_sheet = wb.worksheets[index+1]
                    for index, header in enumerate(sh.tables[0]['dataframe'].columns.names):
                        wb_sheet.row_dimensions[top_offset+index+1].height=heights[header]
                        for col_range in range(1,100):
                            wb_sheet.cell(top_offset+index+1,col_range).alignment = Alignment(wrap_text=True, vertical='center', horizontal='center')

                    for row in wb_sheet.rows:
                        for cell in row:
                            pass
                            #cell.font = default_font                



            else:
                wb_sheet = wb.worksheets[0]
                if self.logo is not None:                
                    img = Image(self.logo)
                    wb_sheet.add_image(img, 'B2')

                for row_range in range(1, self.sheets[0].options.formats['offsets']['top']+1):
                    for col_range in range(1, 30):
                        color_cell = wb_sheet.cell(row_range, col_range)
                        color_cell.fill = PatternFill(start_color=self.index_options['header_color'], end_color=self.index_options['body_color'], fill_type="solid")

                top_offset = self.sheets[0].options.formats['offsets']['top']
                for index, header in enumerate(self.sheets[0].tables[0]['dataframe'].columns.names):
                    wb_sheet.row_dimensions[top_offset+index+1].height=heights[header]
                    for col_range in range(1,100):
                        wb_sheet.cell(top_offset+index+1,col_range).alignment = Alignment(wrap_text=True, vertical='center', horizontal='center')
 

            wb.save(filename)
        except Exception as e:
            print('Warning: Something went wrong in prettifying the Excel. Make sure all colors are defined as ffffff rather than #ffffff.')
            print(e)
            pass

    def table_count(self):
        table_counts = [i.table_count() for i in self.sheets]
        return sum(table_counts)



import functools
import io
import json
import pandas as pd

import tally

def add_data(func):
    @functools.wraps(func)
    def wrapper(*aargs, **kkwargs):
        if aargs[0].dataset_type == 'quantipy':
            kkwargs['data_params'] = {
                'meta': aargs[0].qp_meta, 
                'data': aargs[0].qp_data
            }
        elif aargs[0].dataset_type == 'sav':
            kkwargs['data_params'] = {
                'binary_data': {'spss': (aargs[0].filename, 
                                  io.BytesIO(aargs[0].sav_data), 
                                  'application/x-spss-sav')
                        }
            }
        return func(*aargs, **kkwargs)
    return wrapper

def format_response(func):
    @functools.wraps(func)
    def wrapper(*aargs, **kkwargs):
        format = kkwargs.pop('format', 'dataframe')
        result = func(*aargs, **kkwargs)
        if 'text' in result:
            return result['text']
        if format == 'dict':
            result = result
        elif format == 'json':
            result = json.dumps(result)
        elif format == 'dataframe':
            if 'result' in result.keys():
                result = tally.result_to_dataframe(result['result'])
            elif all(elem in result.keys() for elem in ['index', 'columns', 'data']):
                result = tally.result_to_dataframe(result)
            else:
                if 'params' in result:
                    del result['params']
                result = pd.DataFrame({k:pd.Series(v, dtype='object') for k,v in result.items()})
        else:
            pass # Do nothint, internally use json
        return result
    return wrapper

def valid_params(valid_params_list):
    """
    A decorator that gets list of valid parameters and returns an error if user supplies a parameter
    that isn't in the list. Useful for making sure invalid params aren't sent up to Tally.
    """
    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*aargs, **kkwargs):
            sent_args = list(kkwargs.keys())
            invalid = list(set(sent_args) - set(valid_params_list))
            if len(invalid) > 0:
                raise ValueError("invalid parameter: {}. Valid parameters are: {}\n See https://tally.datasmoothie.com for documentation.".format(invalid, valid_params_list))
            result = func(*aargs, **kkwargs)
            return result
        return wrapper
    return actual_decorator

    
def verify_no_tables(func):
    @functools.wraps(func)
    def wrapper(*aargs, **kkwargs):
        error = f"You are using {func.__name__} after you have used sheet.add_table. Options have to be set before add_table is called"
        if aargs[0].parent.table_count() > 0:
            raise Exception(error)

        result = func(*aargs, **kkwargs)
        return result
    return wrapper



from .decorators import verify_no_tables


class Options:

    def __init__(self, table_options=None, formats={}, parent=None):
        if table_options is None:
            self.table_options = {
                "format":{
                    "base":{},
                    "percentage":{},
                    "counts":{}
                },
                "stub": {"ci":["c%"]},
                "banner_border": True
            }
        else:
            self.table_options = table_options

        self.formats = formats
        self.parent = parent

    def set_base_position(self, position):
        if position == 'outside':
            self.set_question_format('base', {'font_color':'ffffff', 'bg_color':'ffffff', 'font_size':1})
        self.table_options['base'] = position

    def set_language_key(self, text_key):
        self.table_options['stub']['text_key'] = text_key

    def set_show_table_base_column(self, xtotal):
        self.table_options['stub']['xtotal'] = xtotal

    def set_sig_test_levels(self,sig_test_level):
        if type(sig_test_level) == list:
            self.table_options['stub']['sig_level'] = sig_test_level
        else:
            self.table_options['stub']['sig_level'] = [sig_test_level]

    def set_hide_gridlines(self, hide):
        """
        0 Don’t hide gridlines.
        1 Hide printed gridlines only.
        2 Hide screen and printed gridlines.
        """
        self._set_page_setup('hide_gridlines', {"option":hide})

    @verify_no_tables
    def set_stats(self, stats=['mean']):
        self.table_options['stub']['stats'] = stats

    @verify_no_tables
    def set_filter(self, filter):
        self.table_options['stub']['f'] = filter

    @verify_no_tables
    def set_ci(self, ci):
        self.table_options['stub']['ci'] = ci

    @verify_no_tables
    def set_show_bases(self, bases):
        if bases not in ['auto', 'weighted', 'unweighted', 'both']:
            raise ValueError('Bases to show must be both, auto, weighted or unweighted')
        self.table_options['stub']['base'] = bases

    @verify_no_tables
    def set_stub(self, default_stub):
        self.table_options['stub'] = {**self.table_options['stub'], **default_stub}

    @verify_no_tables
    def set_weight(self, weight):
        self.table_options['stub']['w'] = weight

    @verify_no_tables
    def set_banner_border(self, border):
        self.table_options['banner_border'] = border

    def freeze_panes(self, row=10, column=1):
        freeze = {
            "row": row,
            "col": column
        }
        self.formats['freeze_panes'] = freeze


    def set_row_colors(self, colors):
        """
        Params:

        colors - list of colours that will be alternated across (supply to colours to get stripes)
        """
        if type(colors) == list:
            self.formats['row_colors'] = colors
        else:
            raise ValueError("The colours must be a list of one or more colours")

    def set_format(self, name, new_format):
        if name in self.formats:
            self.formats[name] = {**self.formats[name], **new_format}
        else:
            self.formats[name] = new_format

    @verify_no_tables
    def set_answer_format(self, answer_type, format):
        new_format = {0: {"format": format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    def set_question_format(self, answer_type, format):
        new_format = {"question": {"format": format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    @verify_no_tables
    def set_column_format_for_type(self, answer_type, column_index, format):
        new_format = {column_index: {"format":format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    def _set_page_setup(self, key, value):
        self.formats['set_page_setup'][key] = value

    def merge(self,  options):
        """ Merge two Option objects together, this one overriding the incoming one
        """
        self.table_options = self.merge_dict(self.table_options, options.table_options)
        self.formats = self.merge_dict(self.formats, options.formats)

    def merge_dict(self, source, destination):
        """
        >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
        """
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, {})
                self.merge_dict(value, node)
            else:
                destination[key] = value

        return destination


import setuptools


setuptools.setup(
    name="datasmoothie-tally-client",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'openpyxl',
        'pillow'
    ],
    version="0.31",
    license='MIT',
    include_package_data=True,
    url="https://github.com/datasmoothie/datasmoothie-tally-client",
    author="Geir Freysson",
    author_email="geir@datasmoothie.com",
    description="Python wrapper for the Tally API.",
    keywords=['surveys', 'market research', 'weighting', 'significance tests'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
